@@ -1,40 +1,26 @@
-package org.glyptodon.guacamole.servlet;
+/*
+ * Copyright (C) 2015 Glyptodon LLC
+ *
+ * Permission is hereby granted, free of charge, to any person obtaining a copy
+ * of this software and associated documentation files (the "Software"), to deal
+ * in the Software without restriction, including without limitation the rights
+ * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+ * copies of the Software, and to permit persons to whom the Software is
+ * furnished to do so, subject to the following conditions:
+ *
+ * The above copyright notice and this permission notice shall be included in
+ * all copies or substantial portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+ * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+ * THE SOFTWARE.
+ */
 
-/* ***** BEGIN LICENSE BLOCK *****
- * Version: MPL 1.1/GPL 2.0/LGPL 2.1
- *
- * The contents of this file are subject to the Mozilla Public License Version
- * 1.1 (the "License"); you may not use this file except in compliance with
- * the License. You may obtain a copy of the License at
- * http://www.mozilla.org/MPL/
- *
- * Software distributed under the License is distributed on an "AS IS" basis,
- * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
- * for the specific language governing rights and limitations under the
- * License.
- *
- * The Original Code is guacamole-common.
- *
- * The Initial Developer of the Original Code is
- * Michael Jumper.
- * Portions created by the Initial Developer are Copyright (C) 2010
- * the Initial Developer. All Rights Reserved.
- *
- * Contributor(s):
- *
- * Alternatively, the contents of this file may be used under the terms of
- * either the GNU General Public License Version 2 or later (the "GPL"), or
- * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
- * in which case the provisions of the GPL or the LGPL are applicable instead
- * of those above. If you wish to allow use of your version of this file only
- * under the terms of either the GPL or the LGPL, and not to allow others to
- * use your version of this file under the terms of the MPL, indicate your
- * decision by deleting the provisions above and replace them with the notice
- * and other provisions required by the GPL or the LGPL. If you do not delete
- * the provisions above, a recipient may use your version of this file under
- * the terms of any one of the MPL, the GPL or the LGPL.
- *
- * ***** END LICENSE BLOCK ***** */
+package org.glyptodon.guacamole.servlet;
 
 import java.io.BufferedWriter;
 import java.io.IOException;
@@ -46,15 +32,15 @@ import javax.servlet.ServletException;
 import javax.servlet.http.HttpServlet;
 import javax.servlet.http.HttpServletRequest;
 import javax.servlet.http.HttpServletResponse;
-import javax.servlet.http.HttpSession;
 import org.glyptodon.guacamole.GuacamoleClientException;
+import org.glyptodon.guacamole.GuacamoleConnectionClosedException;
 import org.glyptodon.guacamole.GuacamoleException;
 import org.glyptodon.guacamole.GuacamoleResourceNotFoundException;
-import org.glyptodon.guacamole.GuacamoleSecurityException;
 import org.glyptodon.guacamole.GuacamoleServerException;
 import org.glyptodon.guacamole.io.GuacamoleReader;
 import org.glyptodon.guacamole.io.GuacamoleWriter;
 import org.glyptodon.guacamole.net.GuacamoleTunnel;
+import org.glyptodon.guacamole.protocol.GuacamoleStatus;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
@@ -69,7 +55,12 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
     /**
      * Logger for this class.
      */
-    private Logger logger = LoggerFactory.getLogger(GuacamoleHTTPTunnelServlet.class);
+    private final Logger logger = LoggerFactory.getLogger(GuacamoleHTTPTunnelServlet.class);
+
+    /**
+     * Map of absolutely all active tunnels using HTTP, indexed by tunnel UUID.
+     */
+    private final GuacamoleHTTPTunnelMap tunnels = new GuacamoleHTTPTunnelMap();
 
     /**
      * The prefix of the query string which denotes a tunnel read operation.
@@ -96,6 +87,56 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
      */
     private static final int UUID_LENGTH = 36;
 
+    /**
+     * Registers the given tunnel such that future read/write requests to that
+     * tunnel will be properly directed.
+     *
+     * @param tunnel
+     *     The tunnel to register.
+     */
+    protected void registerTunnel(GuacamoleTunnel tunnel) {
+        tunnels.put(tunnel.getUUID().toString(), tunnel);
+        logger.debug("Registered tunnel \"{}\".", tunnel.getUUID());
+    }
+
+    /**
+     * Deregisters the given tunnel such that future read/write requests to
+     * that tunnel will be rejected.
+     *
+     * @param tunnel
+     *     The tunnel to deregister.
+     */
+    protected void deregisterTunnel(GuacamoleTunnel tunnel) {
+        tunnels.remove(tunnel.getUUID().toString());
+        logger.debug("Deregistered tunnel \"{}\".", tunnel.getUUID());
+    }
+
+    /**
+     * Returns the tunnel with the given UUID, if it has been registered with
+     * registerTunnel() and not yet deregistered with deregisterTunnel().
+     *
+     * @param tunnelUUID
+     *     The UUID of registered tunnel.
+     *
+     * @return
+     *     The tunnel corresponding to the given UUID.
+     *
+     * @throws GuacamoleException
+     *     If the requested tunnel does not exist because it has not yet been
+     *     registered or it has been deregistered.
+     */
+    protected GuacamoleTunnel getTunnel(String tunnelUUID)
+            throws GuacamoleException {
+
+        // Pull tunnel from map
+        GuacamoleTunnel tunnel = tunnels.get(tunnelUUID);
+        if (tunnel == null)
+            throw new GuacamoleResourceNotFoundException("No such tunnel.");
+
+        return tunnel;
+
+    }
+
     @Override
     protected void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException {
         handleTunnelRequest(request, response);
@@ -107,21 +148,33 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
     }
 
     /**
-     * Sends an error on the given HTTP response with the given integer error
-     * code.
+     * Sends an error on the given HTTP response using the information within
+     * the given GuacamoleStatus.
      *
-     * @param response The HTTP response to use to send the error.
-     * @param code The HTTP status code of the error.
-     * @throws ServletException If an error prevents sending of the error
-     *                          code.
+     * @param response
+     *     The HTTP response to use to send the error.
+     *
+     * @param guacStatus
+     *     The status to send
+     *
+     * @param message
+     *     A human-readable message that can be presented to the user.
+     *
+     * @throws ServletException
+     *     If an error prevents sending of the error code.
      */
-    private void sendError(HttpServletResponse response, int code) throws ServletException {
+    protected void sendError(HttpServletResponse response,
+            GuacamoleStatus guacStatus, String message)
+            throws ServletException {
 
         try {
 
-            // If response not committed, send error code
-            if (!response.isCommitted())
-                response.sendError(code);
+            // If response not committed, send error code and message
+            if (!response.isCommitted()) {
+                response.addHeader("Guacamole-Status-Code", Integer.toString(guacStatus.getGuacamoleStatusCode()));
+                response.addHeader("Guacamole-Error-Message", message);
+                response.sendError(guacStatus.getHttpStatusCode());
+            }
 
         }
         catch (IOException ioe) {
@@ -134,19 +187,23 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
 
     }
 
-
-
     /**
      * Dispatches every HTTP GET and POST request to the appropriate handler
      * function based on the query string.
      *
-     * @param request The HttpServletRequest associated with the GET or POST
-     *                request received.
-     * @param response The HttpServletResponse associated with the GET or POST
-     *                 request received.
-     * @throws ServletException If an error occurs while servicing the request.
+     * @param request
+     *     The HttpServletRequest associated with the GET or POST request
+     *     received.
+     *
+     * @param response
+     *     The HttpServletResponse associated with the GET or POST request
+     *     received.
+     *
+     * @throws ServletException
+     *     If an error occurs while servicing the request.
      */
-    protected void handleTunnelRequest(HttpServletRequest request, HttpServletResponse response) throws ServletException {
+    protected void handleTunnelRequest(HttpServletRequest request,
+            HttpServletResponse response) throws ServletException {
 
         try {
 
@@ -161,14 +218,8 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
                 GuacamoleTunnel tunnel = doConnect(request);
                 if (tunnel != null) {
 
-                    // Get session
-                    HttpSession httpSession = request.getSession(true);
-                    GuacamoleSession session = new GuacamoleSession(httpSession);
-
-                    // Attach tunnel to session
-                    session.attachTunnel(tunnel);
-
-                    logger.info("Connection from {} succeeded.", request.getRemoteAddr());
+                    // Register newly-created tunnel
+                    registerTunnel(tunnel);
 
                     try {
                         // Ensure buggy browsers do not cache response
@@ -184,10 +235,8 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
                 }
 
                 // Failed to connect
-                else {
-                    logger.info("Connection from {} failed.", request.getRemoteAddr());
+                else
                     throw new GuacamoleResourceNotFoundException("No tunnel created.");
-                }
 
             }
 
@@ -212,69 +261,67 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
 
         // Catch any thrown guacamole exception and attempt to pass within the
         // HTTP response, logging each error appropriately.
-        catch (GuacamoleSecurityException e) {
-            logger.warn("Authorization failed.", e);
-            sendError(response, HttpServletResponse.SC_FORBIDDEN);
-        }
-        catch (GuacamoleResourceNotFoundException e) {
-            logger.debug("Resource not found.", e);
-            sendError(response, HttpServletResponse.SC_NOT_FOUND);
-        }
         catch (GuacamoleClientException e) {
-            logger.warn("Error in client request.", e);
-            sendError(response, HttpServletResponse.SC_BAD_REQUEST);
+            logger.warn("HTTP tunnel request rejected: {}", e.getMessage());
+            sendError(response, e.getStatus(), e.getMessage());
         }
         catch (GuacamoleException e) {
-            logger.error("Server error in tunnel", e);
-            sendError(response, HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
+            logger.error("HTTP tunnel request failed: {}", e.getMessage());
+            logger.debug("Internal error in HTTP tunnel.", e);
+            sendError(response, e.getStatus(), "Internal server error.");
         }
 
     }
 
     /**
      * Called whenever the JavaScript Guacamole client makes a connection
-     * request. It it up to the implementor of this function to define what
-     * conditions must be met for a tunnel to be configured and returned as a
-     * result of this connection request (whether some sort of credentials must
-     * be specified, for example).
+     * request via HTTP. It it up to the implementor of this function to define
+     * what conditions must be met for a tunnel to be configured and returned
+     * as a result of this connection request (whether some sort of credentials
+     * must be specified, for example).
      *
-     * @param request The HttpServletRequest associated with the connection
-     *                request received. Any parameters specified along with
-     *                the connection request can be read from this object.
-     * @return A newly constructed GuacamoleTunnel if successful,
-     *         null otherwise.
-     * @throws GuacamoleException If an error occurs while constructing the
-     *                            GuacamoleTunnel, or if the conditions
-     *                            required for connection are not met.
+     * @param request
+     *     The HttpServletRequest associated with the connection request
+     *     received. Any parameters specified along with the connection request
+     *     can be read from this object.
+     *
+     * @return
+     *     A newly constructed GuacamoleTunnel if successful, null otherwise.
+     *
+     * @throws GuacamoleException
+     *     If an error occurs while constructing the GuacamoleTunnel, or if the
+     *     conditions required for connection are not met.
      */
-    protected abstract GuacamoleTunnel doConnect(HttpServletRequest request) throws GuacamoleException;
+    protected abstract GuacamoleTunnel doConnect(HttpServletRequest request)
+            throws GuacamoleException;
 
     /**
      * Called whenever the JavaScript Guacamole client makes a read request.
      * This function should in general not be overridden, as it already
      * contains a proper implementation of the read operation.
      *
-     * @param request The HttpServletRequest associated with the read request
-     *                received.
-     * @param response The HttpServletResponse associated with the write request
-     *                 received. Any data to be sent to the client in response
-     *                 to the write request should be written to the response
-     *                 body of this HttpServletResponse.
-     * @param tunnelUUID The UUID of the tunnel to read from, as specified in
-     *                   the write request. This tunnel must be attached to
-     *                   the Guacamole session.
-     * @throws GuacamoleException If an error occurs while handling the read
-     *                            request.
+     * @param request
+     *     The HttpServletRequest associated with the read request received.
+     *
+     * @param response
+     *     The HttpServletResponse associated with the write request received.
+     *     Any data to be sent to the client in response to the write request
+     *     should be written to the response body of this HttpServletResponse.
+     *
+     * @param tunnelUUID
+     *     The UUID of the tunnel to read from, as specified in the write
+     *     request. This tunnel must have been created by a previous call to
+     *     doConnect().
+     *
+     * @throws GuacamoleException
+     *     If an error occurs while handling the read request.
      */
-    protected void doRead(HttpServletRequest request, HttpServletResponse response, String tunnelUUID) throws GuacamoleException {
-
-        HttpSession httpSession = request.getSession(false);
-        GuacamoleSession session = new GuacamoleSession(httpSession);
+    protected void doRead(HttpServletRequest request,
+            HttpServletResponse response, String tunnelUUID)
+            throws GuacamoleException {
 
         // Get tunnel, ensure tunnel exists
-        GuacamoleTunnel tunnel = session.getTunnel(tunnelUUID);
-        if (tunnel == null)
-            throw new GuacamoleResourceNotFoundException("No such tunnel.");
+        GuacamoleTunnel tunnel = getTunnel(tunnelUUID);
 
         // Ensure tunnel is open
         if (!tunnel.isOpen())
@@ -298,11 +345,11 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
             // Stream data to response, ensuring output stream is closed
             try {
 
-                // Detach tunnel and throw error if EOF (and we haven't sent any
-                // data yet.
+                // Deregister tunnel and throw error if we reach EOF without
+                // having ever sent any data
                 char[] message = reader.read();
                 if (message == null)
-                    throw new GuacamoleResourceNotFoundException("Tunnel reached end of stream.");
+                    throw new GuacamoleConnectionClosedException("Tunnel reached end of stream.");
 
                 // For all messages, until another stream is ready (we send at least one message)
                 do {
@@ -323,8 +370,10 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
                 } while (tunnel.isOpen() && (message = reader.read()) != null);
 
                 // Close tunnel immediately upon EOF
-                if (message == null)
+                if (message == null) {
+                    deregisterTunnel(tunnel);
                     tunnel.close();
+                }
 
                 // End-of-instructions marker
                 out.write("0.;");
@@ -332,27 +381,42 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
                 response.flushBuffer();
             }
 
+            // Send end-of-stream marker and close tunnel if connection is closed
+            catch (GuacamoleConnectionClosedException e) {
+
+                // Deregister and close
+                deregisterTunnel(tunnel);
+                tunnel.close();
+
+                // End-of-instructions marker
+                out.write("0.;");
+                out.flush();
+                response.flushBuffer();
+
+            }
+
+            catch (GuacamoleException e) {
+
+                // Deregister and close
+                deregisterTunnel(tunnel);
+                tunnel.close();
+
+                throw e;
+            }
+
             // Always close output stream
             finally {
                 out.close();
             }
 
         }
-        catch (GuacamoleException e) {
-
-            // Detach and close
-            session.detachTunnel(tunnel);
-            tunnel.close();
-
-            throw e;
-        }
         catch (IOException e) {
 
             // Log typically frequent I/O error if desired
             logger.debug("Error writing to servlet output stream", e);
 
-            // Detach and close
-            session.detachTunnel(tunnel);
+            // Deregister and close
+            deregisterTunnel(tunnel);
             tunnel.close();
 
         }
@@ -367,25 +431,27 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
      * This function should in general not be overridden, as it already
      * contains a proper implementation of the write operation.
      *
-     * @param request The HttpServletRequest associated with the write request
-     *                received. Any data to be written will be specified within
-     *                the body of this request.
-     * @param response The HttpServletResponse associated with the write request
-     *                 received.
-     * @param tunnelUUID The UUID of the tunnel to write to, as specified in
-     *                   the write request. This tunnel must be attached to
-     *                   the Guacamole session.
-     * @throws GuacamoleException If an error occurs while handling the write
-     *                            request.
+     * @param request
+     *     The HttpServletRequest associated with the write request received.
+     *     Any data to be written will be specified within the body of this
+     *     request.
+     *
+     * @param response
+     *     The HttpServletResponse associated with the write request received.
+     *
+     * @param tunnelUUID
+     *     The UUID of the tunnel to write to, as specified in the write
+     *     request. This tunnel must have been created by a previous call to
+     *     doConnect().
+     *
+     * @throws GuacamoleException
+     *     If an error occurs while handling the write request.
      */
-    protected void doWrite(HttpServletRequest request, HttpServletResponse response, String tunnelUUID) throws GuacamoleException {
+    protected void doWrite(HttpServletRequest request,
+            HttpServletResponse response, String tunnelUUID)
+            throws GuacamoleException {
 
-        HttpSession httpSession = request.getSession(false);
-        GuacamoleSession session = new GuacamoleSession(httpSession);
-
-        GuacamoleTunnel tunnel = session.getTunnel(tunnelUUID);
-        if (tunnel == null)
-            throw new GuacamoleResourceNotFoundException("No such tunnel.");
+        GuacamoleTunnel tunnel = getTunnel(tunnelUUID);
 
         // We still need to set the content type to avoid the default of
         // text/html, as such a content type would cause some browsers to
@@ -426,10 +492,13 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
             }
 
         }
+        catch (GuacamoleConnectionClosedException e) {
+            logger.debug("Connection to guacd closed.", e);
+        }
         catch (IOException e) {
 
-            // Detach and close
-            session.detachTunnel(tunnel);
+            // Deregister and close
+            deregisterTunnel(tunnel);
             tunnel.close();
 
             throw new GuacamoleServerException("I/O Error sending data to server: " + e.getMessage(), e);
@@ -440,6 +509,11 @@ public abstract class GuacamoleHTTPTunnelServlet extends HttpServlet {
 
     }
 
+    @Override
+    public void destroy() {
+        tunnels.shutdown();
+    }
+
 }
 
 /**
